
import audiogen
import audiogen.util

import itertools
import contextlib

HERTZ = 770
BANDWIDTH = 200
DEFAULT_WPM = 20

#
# todo: support parsing of prosigns
#

# ARRL Morse Transmission Timing Standard
# http://www.arrl.org/files/file/Technology/x9004008.pdf

# WPM determined by P-A-R-I-S, or 50 dit lengths (with spaces)
# ms/DIT = (60 / 50) / WPM
# WPM = (60 / 50) / (ms/DIT)
# or 
# s/DIT = 1.2 / WPM

def wpm(wpm=20, farnsworth_limit=18):
	'''
	Generate Morse timings for specified WPM

	For WPM values below farnsworth_limit, Farnsworth timing 
	will be used per the ARRL standard. The default farnsworth_limit
	is 18 WPM. To disable Farnsworth timing entirely, set
	farnsworth_limit=None.
	'''
	return farnsworth(
		wpm, 
		max(farnsworth_limit, wpm) if farnsworth_limit else wpm
	)

def farnsworth(wpm=20, cwpm=None):
	'''
	Generate Morse timings based on ARRL Farnsworth timing

	This permits optionally specifying a different overall words
	per minute. The character WPM determines the timings of the 
	dits, dahs, and intra-letter spaces; the overall WPM determines
	the timings of the inter-letter and inter-word spaces.
	'''
	if not cwpm:
		cwpm = wpm

	# ensure cwpm is never slower than wpm!
	cwpm = max(cwpm, wpm)

	dit = 1.2 / cwpm
	# see ARRL doc
	t_a = (60 * cwpm - 37.2 * wpm) / (cwpm * wpm)
	
	return {
		"dit": dit,
		"dah": dit * 3,
		"inter_symbol": dit,

		"inter_letter": (t_a * 3) / 19.,
		"inter_word": (t_a * 7) / 19.
	}

def set_times(timings):
	global DIT, DAH, INTER_SYMBOL, INTER_LETTER, INTER_WORD
	DIT = timings["dit"]
	DAH = timings["dah"]
	INTER_SYMBOL = timings["inter_symbol"]
	INTER_LETTER = timings["inter_letter"]
	INTER_WORD = timings["inter_word"]

def get_times():
	return {
		"dit": DIT,
		"dah": DAH,
		"inter_symbol": INTER_SYMBOL,
		"inter_letter": INTER_LETTER,
		"inter_word": INTER_WORD
	}

# set default timings
set_times(wpm(DEFAULT_WPM))

class timings(object):
	'''
	Context manager to set Morse symbol timings.
	'''
	def __init__(self, timings):
		self._new_timings = timings
	def __enter__(self, ):
		self._saved_timings = get_times()
		set_times(self._new_timings)
	def __exit__(self, *args, **kwargs):
		set_times(self._saved_timings)

class tone(object):
	'''
	Context manager to set frequency for Morse tones.
	'''
	def __init__(self, frequency):
		self._new_frequency = frequency
	def __enter__(self, ):
		global HERTZ
		self._saved_frequency = HERTZ
		HERTZ = self._new_frequency
	def __exit__(self, *args, **kwargs):
		global HERTZ
		HERTZ = self._saved_frequency

@contextlib.contextmanager
def bandwidth(bandwidth):
	global BANDWIDTH
	old = BANDWIDTH
	BANDWIDTH = bandwidth
	yield
	BANDWIDTH = old

def tone_for(seconds):
	tone = audiogen.crop(audiogen.util.volume(audiogen.tone(HERTZ), -3), seconds)
	return tone

def dit():
	for sample in tone_for(DIT):
		yield sample

def dah():
	for sample in tone_for(DAH):
		yield sample

def space():
	for sample in audiogen.silence(INTER_WORD):
		yield sample

def inter_symbol():
	for sample in audiogen.silence(INTER_SYMBOL):
		yield sample

def inter_letter():
	for sample in audiogen.silence(INTER_LETTER):
		yield sample
	
LETTERS = {
	"A": (dit, dah),
	"B": (dah, dit, dit, dit),
	"C": (dah, dit, dah, dit),
	"D": (dah, dit, dit),
	"E": (dit,),
	"F": (dit, dit, dah, dit),
	"G": (dah, dah, dit),
	"H": (dit, dit, dit, dit),
	"I": (dit, dit),
	"J": (dit, dah, dah, dah),
	"K": (dah, dit, dah),
	"L": (dit, dah, dit, dit),
	"M": (dah, dah),
	"N": (dah, dit),
	"O": (dah, dah, dah),
	"P": (dit, dah, dah, dit),
	"Q": (dah, dah, dit, dah),
	"R": (dit, dah, dit),
	"S": (dit, dit, dit),
	"T": (dah,),
	"U": (dit, dit, dah),
	"V": (dit, dit, dit, dah),
	"W": (dit, dah, dah),
	"X": (dah, dit, dit, dah),
	"Y": (dah, dit, dah, dah),
	"Z": (dah, dah, dit, dit),
	"1": (dit, dah, dah, dah, dah),
	"2": (dit, dit, dah, dah, dah),
	"3": (dit, dit, dit, dah, dah),
	"4": (dit, dit, dit, dit, dah),
	"5": (dit, dit, dit, dit, dit),
	"6": (dah, dit, dit, dit, dit),
	"7": (dah, dah, dit, dit, dit),
	"8": (dah, dah, dah, dit, dit),
	"9": (dah, dah, dah, dah, dit),
	"0": (dah, dah, dah, dah, dah),
	".": (dit, dah, dit, dah, dit, dah),
	",": (dah, dah, dit, dit, dah, dah),
	"?": (dit, dit, dah, dah, dit, dit),
	" ": (space,),
}

### Add prosigns to LETTERS dict
# todo: do we need to do this, or can the parser generate custom prosigns on the fly?
PROSIGNS = [ "BT", "SK", "AR", "BK", "KN", "CL" ]
for prosign in PROSIGNS:
	LETTERS[prosign] = reduce(lambda a, b: a + b, [LETTERS[char] for char in prosign])

@audiogen.sampler.cache_finite_samples
def letter(letter):
	'''Return a generator for the audio samples of the Morse for `letter`.'''
	# todo: decide on behavior when letter not in LETTERS
	tones = [gen() for gen in LETTERS[letter]]
	spaces = [gen() for gen in [inter_symbol] * (len(tones) - 1) + [inter_letter]]
	# todo: compensate for added space following an inter-word space char
	gens = [symbol for pair in zip(tones, spaces) for symbol in pair]

	# define bandpass filter to limit morse tone bandwidth
	bpf = audiogen.filters.band_pass(HERTZ, BANDWIDTH)

	# chain the band pass filter three times to narrow bandwidth
	return bpf(bpf(bpf(
		itertools.chain(*gens)
		)))

def code(text):
	'''Return a generator for audio samples of the Morse for `text`.'''
	# todo: parse prosigns out of text to be encoded
	# todo: string or list of strings passed?
	gens = [letter(l) for l in text]
	return itertools.chain(*gens)

if __name__ == "__main__":
	import sys
	#message = (sys.argv[1] if len(sys.argv) > 1 else "KD2CNQ").upper()
	pass



#!/usr/bin/env python

from setuptools import setup, find_packages

required_modules = [
	'audiogen',
	'PyAudio',
	]

with open("README.rst", "rb") as f:
	readme = f.read()

setup(
	name="koch",
	version="0.0.3",
	description="Koch method Morse code training program",
	author="Christopher H. Casebeer",
	author_email="christopher@chc.name",
	url="https://github.com/casebeer/koch",

	packages=find_packages(exclude='tests'),
	install_requires=required_modules,

	tests_require=["nose"],
	test_suite="nose.collector",

	entry_points={
		"console_scripts": [
			"koch = koch.koch:main"
		]
	},

	long_description=readme,
	classifiers=[
		"Environment :: Console",
		"Topic :: Multimedia :: Sound/Audio",
		"Topic :: Communications :: Ham Radio",
		"Programming Language :: Python :: 2.7",
	]
)




# coding=utf8

import logging
logging.basicConfig(level=logging.WARN)

import audiogen
import morse

import itertools
import random

import sys

# http://www.qsl.net/n1irz/finley.morse.html
# http://www.codepractice.com/learning.html
# K  M  R  S  U  A  P   T  L  O  W  I  .  N  J  E  F  0   Y  ,  V  G  5  /
# Q  9  Z  H  3  8  B   ?  4  2  7  C  1  D  6  X  <BT>   <SK>  <AR>

DEFAULT_LENGTH = 10
# i.e. start with only the letter 'K'
DEFAULT_TRAINING_CHARACTERS = 1
DEFAULT_HERTZ = 770
DEFAULT_BANDWIDTH = 200
DEFAULT_WPM = 20
# Farnsworth timings below 20 WPM  (n.b. ARRL cutoff is 18 WPM)
DEFAULT_FARNSWORTH_CUTOFF = 20

def koch_alphabet(characters=2):
	# todo: parse alphabet from string, handling prosigns properly
	# todo: accept and parse custom alphabet string
	alphabet = [char for char in "KMRSUAPTLOWI.NJEF0Y,VG5/Q9ZH38B?427C1D6X"] +\
		["BT", "SK", "AR"]
	return alphabet[:characters]

def koch(length=20, alphabet=koch_alphabet(2)):
	def insert_spaces(letters):
		# first char can't be a space
		# todo: better word length distribution
		yield next(letters)
		for letter in letters:
			if random.random() < 1. / 4:
				yield u" "
			yield letter

	letters = (random.choice(alphabet) for i in range(length))
	return u"".join(insert_spaces(letters))

def main():
	import argparse

	parser = argparse.ArgumentParser()
	parser.add_argument("-l", "--length",
		type=int,
		default=DEFAULT_LENGTH,
		help="Length of practise message in characters."
	)
	parser.add_argument("-c", "--characters",
		type=int,
		default=DEFAULT_TRAINING_CHARACTERS,
		help="Number of distinct characters to practise."
	)
	parser.add_argument("-i", "--intro",
		action="store_true",
		default=False,
		help="Play just the Nth Koch character to introduce it."
	)
	parser.add_argument("-a", "--custom-alphabet",
		type=str,
		default=None,
		help="Custom alphabet to use in place of default Koch ordering."
	)
	# todo: make saving in lieu of playing?
	parser.add_argument("-f", "--file",
		type=str,
		default=None,
		help="Save audio to a WAV file."
	)
	parser.add_argument("-H", "--hertz",
		type=float,
		default=DEFAULT_HERTZ,
		help="Frequency in Hertz to use for practise tones."
	)
	parser.add_argument("-B", "--bandwidth",
		type=float,
		default=DEFAULT_BANDWIDTH,
		help="Audio bandwidth in Hertz, centered on the tone frequency."
	)
	parser.add_argument("-w", "--wpm",
		type=float,
		default=DEFAULT_WPM,
		help="Morse words per minute."
	)
	parser.add_argument("--cwpm",
		type=float,
		default=None,
		help="""Morse words per minute used to generate characters.
		If unset, defaults to WPM when WPM >= 20, or 20 if WPM < 20.
		"""
	)
	parser.add_argument("--forever",
		action="store_true",
		default=False,
		help="""Loop the message forever (cannot be combined with --file)."""
	)
	parser.add_argument("message", nargs="*", default=None)
	args = parser.parse_args()

	if args.forever and args.file:
		print("Cannot write an infinitely looping audio stream to a file.")
		sys.exit(1)

	if args.custom_alphabet:
		# todo: parse with koch_alphabet function (TBD)
		alphabet = args.custom_alphabet[:args.characters]
	else:
		alphabet = koch_alphabet(args.characters)

	if args.message:
		# play manually specified message
		message = u" ".join(args.message).upper()
	elif args.intro:
		# play the Nth char length times to teach the char
		message = alphabet[-1] * args.length
	else:
		# play random message of chosen chars and length
		message = koch(args.length, alphabet)
	
	if args.cwpm:
		cwpm = args.cwpm
	else:
		# Farnsworth timings below 20 WPM  (n.b. ARRL cutoff is 18 WPM)
		cwpm = max(DEFAULT_FARNSWORTH_CUTOFF, args.wpm)

	if args.intro or args.message:
		print(message)
	else:
		print(u"Testing characters{wpm}:\n{chars}".format(
			wpm=u" (Farnsworth {}/{})".format(int(args.wpm), cwpm) \
				if args.wpm < DEFAULT_FARNSWORTH_CUTOFF and not args.cwpm else u"",
			chars=u"·".join(alphabet)
		))
	
	if args.forever:
		print("Hit ctrl-c to exit")
		
	with morse.timings(morse.farnsworth(args.wpm, cwpm)):
		with morse.tone(args.hertz):
			with morse.bandwidth(args.bandwidth):
				# todo: determine nesting behavior of context managers
				audio = morse.code(message)
				if args.file:
					with open(args.file, "wb") as f:
						audiogen.write_wav(f, audio)
				else:
					try:
						if args.forever:
							audio = itertools.cycle(audio)
						stream = audiogen.sampler.play(
							itertools.chain(morse.code(" "), audio, audiogen.beep()),
							blocking=True
						)
					except KeyboardInterrupt:
						# So further messages don't start with "^C"
						print(u"")

	if not args.intro and not args.message and not args.file:
		raw_input(u"\nHit <enter> to continue...")
		print(u"\n{}".format(message.lower()))

	return 0

if __name__ == "__main__":
	sys.exit(main())

